/*
 *  Copyright (C) 2002-2005 Akira Nukada
 *  Copyright (C) 2002-2005 The SSI Project. All rights reserved.
 *  http://ssi.is.s.u-tokyo.ac.jp
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY.  See the COPYING file for the precise
 *  terms and conditions for the use of this software.
 */

#ifndef _FFTSS_H
#define _FFTSS_H      

typedef void *fftss_plan;
typedef double fftss_complex[2];

#define FFTSS_FORWARD          -1
#define FFTSS_BACKWARD         1

#define FFTSS_MEASURE          0
#define FFTSS_DESTROY_INPUT    (1<<0)
#define FFTSS_UNALIGNED        (1<<1)
#define FFTSS_CONSERVE_MEMORY  (1<<2)
#define FFTSS_EXHAUSTIVE       (1<<3)
#define FFTSS_PRESERVE_INPUT   (1<<4)
#define FFTSS_PATIENT          (1<<5)
#define FFTSS_ESTIMATE         (1<<6)

#define FFTSS_NO_SIMD         (1<<17)

#define FFTSS_VERBOSE         (1<<20)
#define FFTSS_INOUT           (1<<21)

#ifdef __cplusplus
extern "C" {
#endif

  extern void *fftss_malloc(long);
  extern void fftss_free(void *);
  
  extern fftss_plan fftss_plan_dft_1d(long, double *, double *, long, long);
  extern fftss_plan fftss_plan_dft_2d(long, long, long, 
				      double *, double *, long, long);
  extern fftss_plan fftss_plan_dft_3d(long, long, long, long, long,
				      double *, double *, long, long);
  extern void fftss_destroy_plan(fftss_plan);

  extern void fftss_set(fftss_plan, double *, double *);
  extern void fftss_execute(fftss_plan);
  extern void fftss_execute_dft(fftss_plan, double *, double *);
  extern double fftss_get_wtime(void);

  extern int fftss_init_threads(void);
  extern void fftss_cleanup_threads(void);
  extern void fftss_plan_with_nthreads(int);
  
  extern int fftss_test_dft_1d(long, double *, double *, long, long);
  
#ifdef __cplusplus
}
#endif

#endif

