/*
 *  Copyright (C) 2002-2005 The SSI Project. All rights reserved.
 *  http://ssi.is.s.u-tokyo.ac.jp
 *
 *  This software is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY.  See the COPYING file for the precise
 *  terms and conditions for the use of this software.
 */
#include <stdio.h>
#include "libfftss.h"

static double d[4];

static void fftss_try_normal(void)
{
  long i;

  double d0, d1, d2, d3;

  d0 = d[0];
  d1 = d[1];
  d2 = d[2];
  d3 = d[3];
  
  for (i = 0; i < (1<<14); i++)
    d0 = d2 * (d1 * (d0 + d1 - d3));

  if (d0 > 1.0) printf("error in fftss_try_normal\n");
}

static void fftss_try_fma(void)
{
  long i;

  double d0, d1, d2;

  d0 = d[0];
  d1 = d[1];
  d2 = d[3];
  
  for (i = 0; i < (1<<14); i++)
    d0 = (d0 * d1 + d2) * d1 + d2;

  if (d0 > 1.0) printf("error in fftss_try_fma\n");
}

int fftss_test_fma(void)
{
  double t0, t1;
  int available;

  d[0] = 1.0; d[1] = 0.5; d[2] = 2.0; d[3] = 0.5;

  t0 = fftss_get_wtime();
  fftss_try_normal();
  t0 = fftss_get_wtime() - t0;

  t1 = fftss_get_wtime();
  fftss_try_fma();
  t1 = fftss_get_wtime() - t1;

  printf("normal: %lf sec.\tfma: %lf sec.\n", t0, t1);
  available = (t0 > t1 * 1.5);
  printf("FMA instructions are %s.\n", 
	 available ? "available" : "not available");

  return available;
}
